import assert from 'assert';
import { isISOTimeFormat, ZmqObserver, randomInt, get, post, put, del } from '../utils/index.mjs';
import { UID } from '../../lib/utils/index.mjs';
import { checkEntities, createLocation, createSublocation, Server } from './helpers.mjs';
import { ERROR_CODE, ZMQ_ACTIONS, ZMQ_CHANNELS, ZMQ_OBJECTS } from '../../lib/system/constants.mjs';

describe('Trackables Management API', function () {
    const trackableOne = {};
    const trackableTwo = {};
    const trackableThree = {};
    const location = {};
    const sublocation = {};

    before(async function () {
        await Server().init();
        Object.assign(location, await createLocation(`Location ${UID()}`));
        Object.assign(sublocation, await createSublocation(`Sublocation ${UID()}`, location));
    });

    after(async function () {
        await del(`/api/v3/locations/${location.uid}`); // cascade remove of sublocations
    });

    describe('Get Empty Trackables', function () {
        it('should return empty list of trackables', checkEntities('trackables', []));
        it('should return NotFound error on wrong uid', function (done) {
            const invalidUid = UID();
            get(`/api/v3/trackables/${invalidUid}`)
                .expect(404)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        `Trackable with uid ${invalidUid} not found`,
                        'Wrong error message'
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Create trackables', function () {
        it('should fail create trackable with string tag_mount_height', function (done) {
            const body = {
                name: 'name',
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 'ss',
                position: null,
            };
        
            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"tag_mount_height":"NOT_INTEGER"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable without name', function (done) {
            const body = {
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"name":"REQUIRED"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with empty name', function (done) {
            const body = {
                name: '',
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"name":"CANNOT_BE_EMPTY"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with name number', function (done) {
            const body = {
                name: 2,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"name":"NOT_STRING"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with wrong heartbeat random value', function (done) {
            const body = {
                name: 'some_name',
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
                heartbeat: 'ss',
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"heartbeat":"NOT_ALLOWED_VALUE"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with wrong heartbeat empty string', function (done) {
            const body = {
                name: 'some_name',
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
                heartbeat: '',
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"heartbeat":"CANNOT_BE_EMPTY"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with negative position_reporting.interval', function (done) {
            const body = {
                name: 'some_name',
                position_reporting: {
                    interval: -1,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"TOO_LOW"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with wrong position_reporting.interval', function (done) {
            const body = {
                name: 'some_name',
                position_reporting: {
                    interval: 0.1,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"NOT_MULTIPLE_TO 0.25"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable without position_reporting', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"position_reporting":"REQUIRED"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with position_reporting={}', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                tag_mount_height: 1000,
                position_reporting: {},
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"REQUIRED","auto":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable without position_reporting.interval', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                position_reporting: {
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable without position_reporting.auto', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                position_reporting: {
                    interval: 0.25,
                },
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"auto":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with interval not multiple to 0.25', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                position_reporting: {
                    interval: 0.4,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"NOT_MULTIPLE_TO 0.25"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with interval string', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                position_reporting: {
                    interval: '0.25',
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };
        
            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"NOT_DECIMAL"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with invalid heartbeat value', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                heartbeat: 'wrong string value',
                tag_mount_height: 1000,
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"heartbeat":"NOT_ALLOWED_VALUE"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with position={}', function (done) {
            const body = {
                name: `Alex_${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: {},
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        `{"position":{"x":"REQUIRED","y":"REQUIRED","z":"REQUIRED","sublocation_uid":"REQUIRED"}}`,
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable if position has not existed sublocation', function (done) {
            const randomUid = UID();
            const body = {
                name: `Alex_${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: {
                    x: randomInt(),
                    y: randomInt(),
                    z: randomInt(),
                    sublocation_uid: randomUid,
                },
            };

            post('/api/v3/trackables', body)
                .expect(404)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        `Sublocation with uid ${randomUid} not found`,
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable if position has no X', function (done) {
            const body = {
                name: `Alex_${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: {
                    y: randomInt(),
                    z: randomInt(),
                    sublocation_uid: sublocation.uid,
                },
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"position":{"x":"REQUIRED"}}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable if position has no sublocation_uid', function (done) {
            const body = {
                name: `Alex_${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: {
                    x: randomInt(),
                    y: randomInt(),
                    z: randomInt(),
                },
            };

            post('/api/v3/trackables', body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position":{"sublocation_uid":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should create trackable one', function (done) {
            const body = {
                name: `Alex_${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: {
                    x: randomInt(),
                    y: randomInt(),
                    z: randomInt(),
                    sublocation_uid: sublocation.uid,
                },
            };

            post('/api/v3/trackables', body)
                .expect(200)
                .then((response) => {
                    Object.assign(trackableOne, response.body);
                    assert.strictEqual(response.body.name, body.name, `Wrong trackable "name" value`);
                    assert.deepStrictEqual(
                        response.body.position_reporting,
                        body.position_reporting,
                        `Wrong trackable "position_reporting" value`
                    );
                    assert.strictEqual(
                        response.body.tag_mount_height,
                        body.tag_mount_height,
                        `Wrong trackable "tag_mount_height" value`
                    );
                    assert.strictEqual(response.body.heartbeat, 'ENABLED', `Default value is enabled`);
                    assert.strictEqual(response.body.tag_mac, null, `Wrong trackable "tag_mac" value`);
                    assert.deepStrictEqual(response.body.position, body.position, `Wrong trackable "position" value`);
                    assert(isISOTimeFormat(response.body.created_at), `Wrong trackable "created_at" format`);
                    assert(isISOTimeFormat(response.body.updated_at), `Wrong trackable "updated_at" format`);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should create trackable two', function (done) {
            const body = {
                name: `Sam_${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: true,
                },
                tag_mount_height: 1000,
                heartbeat: 'disabled',
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(200)
                .then((response) => {
                    Object.assign(trackableTwo, response.body);
                    assert.strictEqual(response.body.name, body.name, `Wrong trackable "name" value`);
                    assert.deepStrictEqual(
                        response.body.position_reporting,
                        body.position_reporting,
                        `Wrong trackable "position_reporting" value`
                    );
                    assert.strictEqual(
                        response.body.tag_mount_height,
                        body.tag_mount_height,
                        `Wrong trackable "tag_mount_height" value`
                    );
                    assert.strictEqual(response.body.tag_mac, null, `Wrong trackable "tag_mac" value`);
                    assert.strictEqual(response.body.heartbeat, 'DISABLED', `Wrong trackable "heartbeat" value`);
                    assert.deepStrictEqual(response.body.position, body.position, `Wrong trackable "position" value`);
                    assert(isISOTimeFormat(response.body.created_at), `Wrong trackable "created_at" format`);
                    assert(isISOTimeFormat(response.body.updated_at), `Wrong trackable "updated_at" format`);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should create trackable three with the reporting interval equal to 0', function (done) {
            const body = {
                name: `Sam_${UID()}`,
                position_reporting: {
                    interval: 0,
                    auto: true,
                },
                tag_mount_height: 1000,
                heartbeat: 'DISABLED',
                position: null,
            };

            post('/api/v3/trackables', body)
                .expect(200)
                .then((response) => {
                    Object.assign(trackableThree, response.body);
                    assert.strictEqual(response.body.name, body.name, `Wrong trackable "name" value`);
                    assert.deepStrictEqual(
                        response.body.position_reporting,
                        body.position_reporting,
                        `Wrong trackable "position_reporting" value`
                    );
                    assert.strictEqual(
                        response.body.tag_mount_height,
                        body.tag_mount_height,
                        `Wrong trackable "tag_mount_height" value`
                    );
                    assert.strictEqual(response.body.tag_mac, null, `Wrong trackable "tag_mac" value`);
                    assert.strictEqual(response.body.heartbeat, 'DISABLED', `Wrong trackable "heartbeat" value`);
                    assert.deepStrictEqual(response.body.position, body.position, `Wrong trackable "position" value`);
                    assert(isISOTimeFormat(response.body.created_at), `Wrong trackable "created_at" format`);
                    assert(isISOTimeFormat(response.body.updated_at), `Wrong trackable "updated_at" format`);
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Update trackables', function () {
        it('should update trackable name', function (done) {
            const body = { name: `Trackable ${UID()}` };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(response.body.name, body.name, `Wrong trackable "name" value`);
                    assert.deepStrictEqual(
                        response.body.position_reporting,
                        trackableOne.position_reporting,
                        `Wrong trackable "position_reporting" value`
                    );
                    assert.strictEqual(
                        response.body.tag_mount_height,
                        trackableOne.tag_mount_height,
                        `Wrong trackable "tag_mount_height" value`
                    );
                    assert.strictEqual(response.body.tag_mac, null, `Wrong trackable "tag_mac" value`);
                    assert.deepStrictEqual(
                        response.body.position,
                        trackableOne.position,
                        `Wrong trackable "position" value`
                    );
                    assert(isISOTimeFormat(response.body.created_at), `Wrong trackable "created_at" format`);
                    assert(isISOTimeFormat(response.body.updated_at), `Wrong trackable "updated_at" format`);
                    assert(
                        new Date(response.body.updated_at) > new Date(trackableOne.updated_at),
                        `"updated_at" should be changed`
                    );
                    Object.assign(trackableOne, response.body);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should update trackable reporting interval to 0', function (done) {
            const body = { position_reporting: { interval: 0, auto: false } };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(200)
                .then((response) => {
                    assert.deepStrictEqual(
                        response.body.position_reporting,
                        body.position_reporting,
                        `Wrong trackable "position_reporting" value`
                    );
                    Object.assign(trackableOne, response.body);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should update trackable heartbeat prop', function (done) {
            const body = { heartbeat: 'DISABLED' };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(response.body.heartbeat, body.heartbeat.toUpperCase(), `Wrong trackable "heartbeat" value`);
                    assert.deepStrictEqual(
                        response.body.position_reporting,
                        trackableOne.position_reporting,
                        `Wrong trackable "position_reporting" value`
                    );
                    assert.strictEqual(
                        response.body.tag_mount_height,
                        trackableOne.tag_mount_height,
                        `Wrong trackable "tag_mount_height" value`
                    );
                    assert.strictEqual(response.body.tag_mac, null, `Wrong trackable "tag_mac" value`);
                    assert.deepStrictEqual(
                        response.body.position,
                        trackableOne.position,
                        `Wrong trackable "position" value`
                    );
                    assert(isISOTimeFormat(response.body.created_at), `Wrong trackable "created_at" format`);
                    assert(isISOTimeFormat(response.body.updated_at), `Wrong trackable "updated_at" format`);
                    assert(
                        new Date(response.body.updated_at) > new Date(trackableOne.updated_at),
                        `"updated_at" should be changed`
                    );
                    Object.assign(trackableOne, response.body);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should update trackable with empty name', function (done) {
            const body = { name: '' };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, `{"name":"CANNOT_BE_EMPTY"}`, `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable with position_reporting={}', function (done) {
            const body = { position_reporting: {} };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"REQUIRED","auto":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable with negative position_reporting.interval', function (done) {
            const body = {
                position_reporting: {
                    interval: -1,
                    auto: false,
                },
            };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"TOO_LOW"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable without position_reporting.interval', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                position_reporting: {
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable without position_reporting.auto', function (done) {
            const body = {
                name: `Trackable ${UID()}`,
                position_reporting: {
                    interval: 0.25,
                },
                tag_mount_height: 1000,
                position: null,
            };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"auto":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable with heartbeat invalid value', function (done) {
            const body = { heartbeat: 'wrong string value' };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"heartbeat":"NOT_ALLOWED_VALUE"}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable with interval invalid value', function (done) {
            const body = { position_reporting: { interval: 0.9 } };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position_reporting":{"interval":"NOT_MULTIPLE_TO 0.25","auto":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable on wrong uid', function (done) {
            const body = { position_reporting: { interval: 0.5, auto: false } };
            const uid = UID();
            put(`/api/v3/trackables/${uid}`, body)
                .expect(404)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, `Trackable with uid ${uid} not found`, `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable with position={}', function (done) {
            const body = {
                name: `Alex_${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: {},
            };
            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        `{"position":{"x":"REQUIRED","y":"REQUIRED","z":"REQUIRED","sublocation_uid":"REQUIRED"}}`,
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable if position has not existed sublocation', function (done) {
            const randomUid = UID();
            const body = {
                position: {
                    x: randomInt(),
                    y: randomInt(),
                    z: randomInt(),
                    sublocation_uid: randomUid,
                },
            };

            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(404)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        `Sublocation with uid ${randomUid} not found`,
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail update trackable if position has no X', function (done) {
            const body = {
                position: {
                    y: randomInt(),
                    z: randomInt(),
                    sublocation_uid: sublocation.uid,
                },
            };

            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(message, '{"position":{"x":"REQUIRED"}}', `Wrong error message`);
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should fail create trackable if position has no sublocation_uid', function (done) {
            const body = {
                position: {
                    x: randomInt(),
                    y: randomInt(),
                    z: randomInt(),
                },
            };

            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(422)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        '{"position":{"sublocation_uid":"REQUIRED"}}',
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.VALIDATION_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should update trackable one position', function (done) {
            const body = {
                position: {
                    x: randomInt(),
                    y: randomInt(),
                    z: randomInt(),
                    sublocation_uid: sublocation.uid,
                },
            };

            put(`/api/v3/trackables/${trackableOne.uid}`, body)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(response.body.name, trackableOne.name, `Wrong trackable "name" value`);
                    assert.deepStrictEqual(
                        response.body.position_reporting,
                        trackableOne.position_reporting,
                        `Wrong trackable "position_reporting" value`
                    );
                    assert.strictEqual(
                        response.body.tag_mount_height,
                        trackableOne.tag_mount_height,
                        `Wrong trackable "tag_mount_height" value`
                    );
                    assert.strictEqual(
                        response.body.heartbeat,
                        trackableOne.heartbeat,
                        `heartbeat must be ${trackableOne.heartbeat}`
                    );
                    assert.strictEqual(response.body.tag_mac, trackableOne.tag_mac, `Wrong trackable "tag_mac" value`);
                    assert.deepStrictEqual(response.body.position, body.position, `Wrong trackable "position" value`);
                    assert(isISOTimeFormat(response.body.created_at), `Wrong trackable "created_at" format`);
                    assert(isISOTimeFormat(response.body.updated_at), `Wrong trackable "updated_at" format`);
                    Object.assign(trackableOne, response.body);
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Read trackables', function () {
        it('should return trackable one', function (done) {
            get(`/api/v3/trackables/${trackableOne.uid}`)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(response.body.trackables.length, 1);
                    assert.deepStrictEqual(response.body.trackables[0], trackableOne, 'trackables not equal');
                    done();
                })
                .catch((err) => done(err));
        });
        it('should return trackables list', function (done) {
            get('/api/v3/trackables')
                .expect(200)
                .then((response) => {
                    const trackables = [trackableOne, trackableTwo];
                    assert.strictEqual(response.body.trackables.length, 3);
                    for (const trackable of trackables) {
                        const resTrackable = response.body.trackables.find((tko) => tko.uid === trackable.uid);
                        assert.deepStrictEqual(resTrackable, trackable, 'trackables not equal');
                    }
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Delete trackables', function () {
        it('should delete trackable one', function (done) {
            del(`/api/v3/trackables/${trackableOne.uid}`)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(response.body.uid, trackableOne.uid, `Wrong trackable "uid" value`);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should delete trackable two', function (done) {
            del(`/api/v3/trackables/${trackableTwo.uid}`)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(response.body.uid, trackableTwo.uid, `Wrong trackable "uid" value`);
                    done();
                })
                .catch((err) => done(err));
        });
        it('should delete trackable three', function (done) {
            del(`/api/v3/trackables/${trackableThree.uid}`)
                .expect(200)
                .then((response) => {
                    assert.strictEqual(response.body.uid, trackableThree.uid, `Wrong trackable "uid" value`);
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('Get Empty Trackables. After removing', function () {
        it('should return empty list of trackables', checkEntities('trackables', []));
        it('should return NotFound error on wrong uid', function (done) {
            const invalidUid = UID();
            get(`/api/v3/trackables/${invalidUid}`)
                .expect(404)
                .then((response) => {
                    const { message, code } = response.body;
                    assert.deepStrictEqual(
                        message,
                        `Trackable with uid ${invalidUid} not found`,
                        `Wrong error message`
                    );
                    assert.deepStrictEqual(code, ERROR_CODE.NOT_FOUND_ERROR, 'Wrong error code');
                    done();
                })
                .catch((err) => done(err));
        });
    });

    describe('ZMQ Messaging for Trackable CRUD', function () {
        const zmqObserver = new ZmqObserver();
        const trackable = {};

        before(function () {
            zmqObserver.init();
        });
        after(function () {
            zmqObserver.destroy();
        });

        it('should send zmq created trackable message', function (done) {
            const zmqSubscription = { action: ZMQ_ACTIONS.CREATED, object: ZMQ_OBJECTS.TRACKABLE };
            const body = {
                name: `Trackable_${UID()}`,
                position_reporting: {
                    interval: 0.25,
                    auto: false,
                },
                tag_mount_height: 1000,
                position: null,
            };

            const subscription = zmqObserver.subscribe(zmqSubscription, function (zmqMsg, channel) {
                assert(isISOTimeFormat(zmqMsg.ts), `Wrong ZMQ message "ts" format`);
                assert(channel === ZMQ_CHANNELS.EVENT, 'Wrong ZMQ message channel');
                assert(zmqMsg.data.uid, 'ZMQ message has no uid');
                assert(zmqMsg.data.name === body.name, 'ZMQ message has wrong trackable name');
                assert(
                    zmqMsg.data.tag_mount_height === body.tag_mount_height,
                    'ZMQ message has wrong tag_mount_height'
                );
                assert(isISOTimeFormat(zmqMsg.data.created_at), 'Wrong ZMQ message created_at value');
                assert(isISOTimeFormat(zmqMsg.data.updated_at), 'Wrong ZMQ message updated_at value');
                // assert(zmqMsg.data.node_mac === null, 'Wrong ZMQ message node_mac value');
                assert.deepStrictEqual(zmqMsg.data.position, body.position, 'Wrong ZMQ message position');
                done();
            });
            subscription
                .then(() => post(`/api/v3/trackables`, body))
                .then((response) => {
                    Object.assign(trackable, response.body);
                });
        });

        it('should send zmq updated trackable message', function (done) {
            const zmqSubscription = { action: ZMQ_ACTIONS.UPDATED, object: ZMQ_OBJECTS.TRACKABLE };
            const body = {
                name: `Trackable_${UID()}`,
            };

            const subscription = zmqObserver.subscribe(zmqSubscription, function (zmqMsg, channel) {
                assert(isISOTimeFormat(zmqMsg.ts), `Wrong ZMQ message "ts" format`);
                assert(channel === ZMQ_CHANNELS.EVENT, 'Wrong ZMQ message channel');
                assert(zmqMsg.data.uid, 'ZMQ message has no uid');
                assert(zmqMsg.data.name === body.name, 'ZMQ message has wrong trackable name');
                assert(
                    zmqMsg.data.tag_mount_height === trackable.tag_mount_height,
                    'ZMQ message has wrong tag_mount_height'
                );
                assert(isISOTimeFormat(zmqMsg.data.created_at), 'Wrong ZMQ message created_at value');
                assert(isISOTimeFormat(zmqMsg.data.updated_at), 'Wrong ZMQ message updated_at value');
                // assert(zmqMsg.data.node_mac === null, 'Wrong ZMQ message node_mac value');
                assert.deepStrictEqual(zmqMsg.data.position, trackable.position, 'Wrong ZMQ message position');
                done();
            });
            subscription.then(() => put(`/api/v3/trackables/${trackable.uid}`, body));
        });
        it('should send zmq deleted trackable message', function (done) {
            const zmqSubscription = { action: ZMQ_ACTIONS.DELETED, object: ZMQ_OBJECTS.TRACKABLE };

            const subscription = zmqObserver.subscribe(zmqSubscription, function (zmqMsg, channel) {
                assert(isISOTimeFormat(zmqMsg.ts), `Wrong ZMQ message "ts" format`);
                assert(channel === ZMQ_CHANNELS.EVENT, 'Wrong ZMQ message channel');
                assert.deepStrictEqual(zmqMsg.data, { uid: trackable.uid }, 'ZMQ message data wrong');
                done();
            });
            subscription.then(() => del(`/api/v3/trackables/${trackable.uid}`));
        });
    });
});
