import makeRequestHandler from '../../../utils/makeRequestHandler.mjs';
import { LocationGet, LocationSave, LocationUpdate, LocationDelete } from '../../../../usecases/index.mjs';

export default {
    /**
     * @swagger
     * /api/v3/locations:
     *   get:
     *     description: Retrieve locations list
     *     tags:
     *       - Locations API
     *     responses:
     *       '200':
     *         description: locations
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/Locations'
     *              example:
     *                  locations:
     *                      -
     *                          uid: oVi39JDNTLO9qdeNM-Ecvg
     *                          name: WestSide
     *                          network_id: 1
     *                          created_at: 2021-04-21T10:00:00.000Z
     *                          updated_at: 2021-04-21T10:00:00.000Z
     *                      -
     *                          uid: oVi39JDNTLO9qdeNM-4444
     *                          name: WestSide2
     *                          network_id: 1
     *                          created_at: 2021-04-21T10:00:00.000Z
     *                          updated_at: 2021-04-21T10:00:00.000Z
     */
    getAll: makeRequestHandler(
        LocationGet,
        () => ({}),
        (result, res) => {
            res.json(result);
        }
    ),
    /**
     * @swagger
     * /api/v3/locations/{uid}:
     *   get:
     *     description: Retrieve single location
     *     tags:
     *       - Locations API
     *     parameters:
     *      - in: path
     *        name: uid
     *        schema:
     *          type: string
     *          required: true
     *          example: oVi39JDNTLO9qdeNM-Ecvg
     *          description: location unique identifier
     *     responses:
     *       '200':
     *         description: locations
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/Locations'
     *              example:
     *                  locations:
     *                      -
     *                          uid: oVi39JDNTLO9qdeNM-Ecvg
     *                          name: WestSide
     *                          network_id: 1
     *                          created_at: 2021-04-21T10:00:00.000Z
     *                          updated_at: 2021-04-21T10:00:00.000Z
     *       '404':
     *         description: Location with the specified UID was not found.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NotFoundError'
     *              example:
     *                  message: Location with uid ... not found
     *                  code: NOT_FOUND_ERROR
     */
    get: makeRequestHandler(
        LocationGet,
        (req) => ({
            uid: req.params.uid,
        }),
        (result, res) => {
            res.json(result);
        }
    ),
    /**
     * @swagger
     * /api/v3/locations:
     *   post:
     *     description: Create location
     *     tags:
     *       - Locations API
     *     requestBody:
     *      content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/LocationCreateBody'
     *              example:
     *                  name: WestSide
     *     responses:
     *       '200':
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/Location'
     *              example:
     *                  uid: oVi39JDNTLO9qdeNM-Ecvg
     *                  name: WestSide
     *                  network_id: 1
     *                  created_at: 2021-04-21T10:00:00.000Z
     *                  updated_at: 2021-04-21T10:00:00.000Z
     *       '422':
     *         description: Invalid request parameters.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/ValidationError'
     *              examples:
     *                  No name specified:
     *                      value:
     *                          message: '{"name":"REQUIRED"}'
     *                          code: VALIDATION_ERROR
     *                  Empty name:
     *                      value:
     *                          message: '{"name":"CANNOT_BE_EMPTY"}'
     *                          code: VALIDATION_ERROR
     */
    post: makeRequestHandler(
        LocationSave,
        (req) => ({
            name: req.body.name,
        }),
        (result, res) => {
            res.json(result);
        }
    ),
    /**
     * @swagger
     * /api/v3/locations/{uid}:
     *   put:
     *     description: Update location
     *     tags:
     *       - Locations API
     *     parameters:
     *      - in: path
     *        name: uid
     *        schema:
     *          type: string
     *          required: true
     *          example: eJEG2Q2kRnSHJaBdHxQ3Pw
     *          description: location unique identifier
     *     requestBody:
     *      content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/LocationUpdateBody'
     *              example:
     *                  name: NorthWest
     *     responses:
     *       '200':
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/Location'
     *              example:
     *                  uid: oVi39JDNTLO9qdeNM-Ecvg
     *                  name: NorthWest
     *                  network_id: 1
     *                  created_at: 2021-04-21T10:00:00.000Z
     *                  updated_at: 2021-04-21T10:00:00.000Z
     *       '404':
     *         description: Location with the specified UID was not found.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NotFoundError'
     *              example:
     *                  message: Location with uid ... not found
     *                  code: NOT_FOUND_ERROR
     *       '422':
     *         description: Invalid request parameters.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/ValidationError'
     *              examples:
     *                  Empty name:
     *                      value:
     *                          message: '{"name":"CANNOT_BE_EMPTY"}'
     *                          code: VALIDATION_ERROR
     */
    put: makeRequestHandler(
        LocationUpdate,
        (req) => ({
            uid: req.params.uid,
            name: req.body.name,
        }),
        (result, res) => {
            res.json(result);
        }
    ),
    /**
     * @swagger
     * /api/v3/locations/{uid}:
     *   delete:
     *     description: Remove location by uid
     *     tags:
     *       - Locations API
     *     parameters:
     *      - in: path
     *        name: uid
     *        schema:
     *          type: string
     *          required: true
     *          example: oVi39JDNTLO9qdeNM-Ecvg
     *          description: location unique identifier
     *     responses:
     *       '200':
     *         content:
     *          application/json:
     *              schema:
     *                  type: object
     *                  properties:
     *                      uid:
     *                          type: string
     *              example:
     *                  uid: oVi39JDNTLO9qdeNM-Ecvg
     *       '404':
     *         description: Location with the specified UID was not found.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NotFoundError'
     *              example:
     *                  message: Location with uid ... not found
     *                  code: NOT_FOUND_ERROR
     */
    del: makeRequestHandler(
        LocationDelete,
        (req) => ({
            uid: req.params && req.params.uid,
        }),
        (result, res) => {
            res.json(result);
        }
    ),
};
