import makeRequestHandler from '../../../utils/makeRequestHandler.mjs';
import { NodeAssociationGet, NodeAssociationSave, NodeAssociationClose } from '../../../../usecases/index.mjs';

export default {
    /**
     * @swagger
     * /api/v3/node_assignments:
     *   get:
     *     description: Retrieve node assignments list
     *     tags:
     *       - Node Assignment API
     *     responses:
     *       '200':
     *         description: node assignments
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NodeAssignments'
     *              example:
     *                  node_assignments:
     *                      -
     *                          anchor_uid: 14bOWPAMSgilkQEb-2wLqg
     *                          mac: E4956EA5010A
     *                          created_at: 2021-04-21T10:00:00.000Z
     *                      -
     *                          anchor_uid: 14bOWPAMSgilkQEb-dfgqR
     *                          mac: E4956EA5010B
     *                          created_at: 2021-04-21T10:00:00.000Z
     *       '404':
     *          description: Invalid request parameters.
     *          content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/ValidationError'
     *              examples:
     *                  No anchor:
     *                      message: Assignment with anchor ... not found
     *                      code: NOT_FOUND_ERROR
     */
    getAll: makeRequestHandler(
        NodeAssociationGet,
        () => ({}),
        (nodeAssociations, res) => {
            res.json({ node_assignments: nodeAssociations });
        }
    ),

    /**
     * @swagger
     * /api/v3/node_assignments:
     *   post:
     *     description: Create node assignment
     *     tags:
     *       - Node Assignment API
     *     requestBody:
     *        content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NodeAssociationCreateBody'
     *              example:
     *                  mac: E4956EA5010A
     *                  anchor_uid: 14bOWPAMSgilkQEb-2wLqg
     *     responses:
     *       '200':
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NodeAssignments'
     *              example:
     *                  anchor_uid: 14bOWPAMSgilkQEb-2wLqg
     *                  mac: E4956EA5010A
     *                  created_at: 2021-04-21T10:00:00.000Z
     *       '422':
     *          description: Invalid request parameters.
     *          content:
     *              application/json:
     *                  schema:
     *                      $ref: '#/components/schemas/ValidationError'
     *                  examples:
     *                      No mac:
     *                          value:
     *                              message: '{"mac":"REQUIRED"}'
     *                              code: VALIDATION_ERROR
     *                      Empty mac:
     *                          value:
     *                              message: '{"mac":"CANNOT_BE_EMPTY"}'
     *                              code: VALIDATION_ERROR
     *                      Wrong mac format:
     *                          value:
     *                              message: '{"mac":"WRONG_MAC_FORMAT"}'
     *                              code: VALIDATION_ERROR
     *                      No anchor_uid:
     *                          value:
     *                              message: '{"anchor_uid":"REQUIRED"}'
     *                              code: VALIDATION_ERROR
     *                      Empty anchor_uid:
     *                          value:
     *                              message: '{"anchor_uid":"CANNOT_BE_EMPTY"}'
     *                              code: VALIDATION_ERROR
     *                      Node assignment exist:
     *                          value:
     *                              message: 'Assignment with node ... already exist'
     *                              code: VALIDATION_ERROR
     *                      Anchor assignment exist:
     *                          value:
     *                              message: 'Assignment with anchor 14bOWPAMSgilkQEb-2wLqg already exist'
     *                              code: VALIDATION_ERROR
     *       '404':
     *         description: Node or anchor with the specified parameters was not found.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NotFoundError'
     *              examples:
     *                  Node not found:
     *                      value:
     *                          message: Node with mac ... not found
     *                          code: NOT_FOUND_ERROR
     *                  Anchor not found:
     *                      value:
     *                          message: Anchor with uid ... not found
     *                          code: NOT_FOUND_ERROR
     */
    post: makeRequestHandler(
        NodeAssociationSave,
        (req) => ({
            mac: req.body.mac,
            anchor_uid: req.body.anchor_uid,
        }),
        (result, res) => {
            res.json(result);
        }
    ),
    /**
     * @swagger
     * /api/v3/node_assignments:
     *   put:
     *     description: Remove node assignment
     *     tags:
     *       - Node Assignment API
     *     requestBody:
     *        content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NodeAssociationCloseBody'
     *              example:
     *                  mac: E4956EA5010A
     *     responses:
     *       '200':
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NodeAssignments'
     *              example:
     *                  anchor_uid: 14bOWPAMSgilkQEb-2wLqg
     *                  mac: E4956EA5010A
     *                  created_at: 2021-04-21T10:00:00.000Z
     *       '422':
     *          description: Invalid request parameters.
     *          content:
     *              application/json:
     *                  schema:
     *                      $ref: '#/components/schemas/ValidationError'
     *                  examples:
     *                      No mac or anchor uid:
     *                          value:
     *                              message: '{"mac":"REQUIRED","anchor_uid":"REQUIRED"}'
     *                              code: VALIDATION_ERROR
     *                      Empty mac:
     *                          value:
     *                              message: '{"mac":"CANNOT_BE_EMPTY"}'
     *                              code: VALIDATION_ERROR
     *                      Empty anchor_uid:
     *                          value:
     *                              message: '{"anchor_uid":"CANNOT_BE_EMPTY"}'
     *                              code: VALIDATION_ERROR
     *                      Wrong mac format:
     *                          value:
     *                              message: '{"mac":"WRONG_MAC_FORMAT"}'
     *                              code: VALIDATION_ERROR
     *       '404':
     *         description: Node assignment with the specified mac address was not found.
     *         content:
     *          application/json:
     *              schema:
     *                  $ref: '#/components/schemas/NotFoundError'
     *              examples:
     *                  Wrong mac and anchor uid:
     *                      message: Assignment with node ... and anchor ... not found
     *                      code: NOT_FOUND_ERROR
     *                  Wrong mac:
     *                      message: Assignment with node ... not found
     *                      code: NOT_FOUND_ERROR
     *                  Wrong anchor uid:
     *                      message: Assignment with anchor ... not found
     *                      code: NOT_FOUND_ERROR
     */
    put: makeRequestHandler(
        NodeAssociationClose,
        (req) => ({
            mac: req.body && req.body.mac,
            anchor_uid: req.body.anchor_uid,
        }),
        (result, res) => {
            res.json(result);
        }
    ),
};
