import ModelBase from '../ModelBase.mjs';
import { NetworkId } from '../id/index.mjs';
import { NotFoundError } from '../errors/index.mjs';
import { UID } from '../../utils/index.mjs';
import { formatLocationResponse, formatLocationResponseList } from './formatters.mjs';

/**
 * @swagger
 * components:
 *  schemas:
 *      LocationCreateBody:
 *          type: object
 *          properties:
 *              name:
 *                  type: string
 *                  required: true
 *      LocationUpdateBody:
 *          type: object
 *          properties:
 *              name:
 *                  type: string
 *                  required: false
 *      Location:
 *          type: object
 *          properties:
 *              uid:
 *                  type: string
 *                  description: Unique id of location
 *              network_id:
 *                  type: number
 *                  description: Network unique numeric (1…65535) identifier
 *              name:
 *                  type: string
 *                  description: Location name
 *              created_at:
 *                  type: string (ISO 8601)
 *                  description: Timestamp of location creation
 *              updated_at:
 *                  type: string (ISO 8601)
 *                  description: Timestamp of location update
 *      Locations:
 *          type: object
 *          properties:
 *              locations:
 *                  type: array
 *                  items:
 *                      $ref: '#/components/schemas/Location'
 */

export class Location extends ModelBase {
    /**
     * @property {String|null}
     */
    static object = 'location';

    /**
     * @method
     * @param {Object} params - location properties to save
     * @param {String} params.name - location name
     * @return {Promise<Object>}
     */
    async save(params) {
        const now = new Date().toISOString();
        const network_id = await new NetworkId().nextValue();

        const location = {
            uid: UID(),
            network_id,
            created_at: now,
            updated_at: now,
            ...params,
        };
        const result = await this.repository.location.save(location);

        return formatLocationResponse(result);
    }

    /**
     @typedef LocationWithChildren
     @type {Object}
     @property {Object} location - location.
     @property {Object} sublocations - The list of sublocation.
     @property {Array<Object>} anchors - The list of anchors.
     @property {Array<Object>} nodeAssociations - The list of node associations.
     */
    /**
     * @method
     * @param {Object} params - options for read location in repository
     * @param {String} params.uid - location unique identifier
     * @return {Promise<LocationWithChildren>}
     */
    async loadWithChildren(params) {
        const location = await this.repository.location.findByUid(params.uid);
        if (!location) {
            throw new NotFoundError(`Location with uid ${params.uid} not found`);
        }
        const sublocations = ((await this.repository.sublocation.findAll()) || []).filter(
            (sublocation) => sublocation.location_uid === params.uid
        );
        const nodeAssociations = [];
        const anchors = ((await this.repository.anchor.findAll()) || []).filter((anchor) =>
            sublocations.find((sublocation) => anchor.position && anchor.position.sublocation_uid === sublocation.uid)
        );
        const networkLsbSchedules = ((await this.repository.networkLsbSchedule.findAll()) || []).filter(
            (schedule) => schedule.location_uid === params.uid
        );

        const networkSetting = await this.repository.networkSetting.findByUid(params.uid);

        for (const anchor of anchors) {
            const nodeAssociation = await this.repository.association.findByObjUid(anchor.uid);
            if (nodeAssociation) nodeAssociations.push(nodeAssociation);
        }

        return { location, sublocations, anchors, nodeAssociations, networkLsbSchedules, networkSetting };
    }

    /**
     * @method
     * @param {Object} params - options for read location in repository
     * @param {String|undefined} params.uid - location unique identifier
     * @return {Promise<Object>}
     */
    async load(params) {
        const { uid } = params;
        if (!uid) {
            const locations = (await this.repository.location.findAll()) || [];
            return { locations: formatLocationResponseList(locations) };
        }
        const location = await this.repository.location.findByUid(uid);
        if (!location) {
            throw new NotFoundError(`Location with uid ${uid} not found`);
        }
        return { locations: [formatLocationResponse(location)] };
    }

    /**
     * @method
     * @param {Object} params - options for read location in repository
     * @param {String} params.uid - location unique identifier
     * @return {Promise<Object>}
     */
    async remove(params) {
        const { uid } = params;
        const location = await this.repository.location.findByUid(uid);
        if (!location) {
            throw new NotFoundError(`Location with uid ${uid} not found`);
        }
        await this.repository.location.deleteByUid(uid);
        await new NetworkId().remove(location.network_id);

        return location;
    }

    /**
     * @method
     * @param {Object} params - location properties to update
     * @param {String} params.uid - location unique identifier
     * @param {String|undefined} params.name - location name
     * @return {Promise<Object>}
     */
    async update(params) {
        const { uid } = params;
        const location = await this.repository.location.findByUid(uid);
        if (!location) {
            throw new NotFoundError(`Location with uid ${uid} not found`);
        }

        await this.repository.location.save({ ...params, updated_at: new Date().toISOString() });

        return formatLocationResponse(await this.repository.location.findByUid(uid));
    }
}
