import ModelBase from '../ModelBase.mjs';
import { SublocationId } from '../id/index.mjs';
import { NotFoundError } from '../errors/index.mjs';
import { UID } from '../../utils/index.mjs';
import { formatSublocationResponse, formatSublocationResponseList } from './formatters.mjs';

/**
 * @swagger
 * components:
 *  schemas:
 *      SublocationCreateBody:
 *          type: object
 *          properties:
 *              name:
 *                  type: string
 *                  required: true
 *              location_uid:
 *                  type: string
 *                  required: true
 *              elevation:
 *                  type: number
 *                  required: false
 *                  default: 0
 *      SublocationUpdateBody:
 *          type: object
 *          properties:
 *              name:
 *                  type: string
 *                  required: false
 *              elevation:
 *                  type: number
 *                  required: false
 *      Sublocation:
 *          type: object
 *          properties:
 *              uid:
 *                  type: string
 *                  description: Sublocation unique identifier.
 *              id:
 *                  type: number
 *                  description: Sublocation numeric (1…65535) identifier. Unique within the location.
 *              name:
 *                  type: string
 *                  description: Sublocation name.
 *              elevation:
 *                  type: number
 *                  description: The sublocation Z offset (elevation) in mm. Multiple to 10.
 *              location_uid:
 *                  type: string
 *                  description: Unique identifier of the parent location.
 *              created_at:
 *                  type: string (ISO 8601)
 *                  description: The time stamp when the sublocation was created, in ISO 8601 format.
 *              updated_at:
 *                  type: string (ISO 8601)
 *                  description: The time stamp when the sublocation was updated, in ISO 8601 format.
 *      Sublocations:
 *          type: object
 *          properties:
 *              sublocations:
 *                  type: array
 *                  items:
 *                      $ref: '#/components/schemas/Sublocation'
 */

export class Sublocation extends ModelBase {
    /**
     * @property {String|null}
     */
    static object = 'sublocation';

    /**
     * @method
     * @param {Object} params - sublocation properties to save
     * @param {String} params.name - sublocation name
     * @param {String} params.location_uid - unique identifier of the parent location
     * @return {Promise<Object>}
     */
    async save(params) {
        const parentLocation = await this.repository.location.findByUid(params.location_uid);

        if (!parentLocation) {
            throw new NotFoundError(`Location with uid ${params.location_uid} not found`);
        }

        const now = new Date().toISOString();
        const id = await new SublocationId().nextValue();

        const sublocation = {
            uid: UID(),
            id,
            created_at: now,
            updated_at: now,
            ...params,
        };
        const result = await this.repository.sublocation.save(sublocation);

        return formatSublocationResponse(result);
    }

    /**
     * @method
     * @param {Object} params - options for read sublocation in repository
     * @param {String|undefined} params.uid - sublocation unique identifier
     * @return {Promise<Object>}
     */
    async load(params) {
        const { uid } = params;
        if (!uid) {
            const sublocations = (await this.repository.sublocation.findAll()) || [];
            return { sublocations: formatSublocationResponseList(sublocations) };
        }
        const sublocation = await this.repository.sublocation.findByUid(uid);
        if (!sublocation) {
            throw new NotFoundError(`Sublocation with uid ${uid} not found`);
        }
        return { sublocations: [formatSublocationResponse(sublocation)] };
    }

    /**
     @typedef SublocationWithChildren
     @type {Object}
     @property {Object} sublocation - sublocation.
     @property {Array<Object>} anchors - The list of anchors.
     @property {Array<Object>} nodeAssociations - The list of node associations.
     */
    /**
     * @method
     * @param {Object} params - options for read sublocation in repository
     * @param {String} params.uid - sublocation unique identifier
     * @return {Promise<SublocationWithChildren>}
     */
    async loadWithChildren(params) {
        const sublocation = await this.repository.sublocation.findByUid(params.uid);
        if (!sublocation) {
            throw new NotFoundError(`Sublocation with uid ${params.uid} not found`);
        }
        const nodeAssociations = [];
        const anchors = ((await this.repository.anchor.findAll()) || []).filter(
            (anchor) => anchor.position && anchor.position.sublocation_uid === params.uid
        );

        for (const anchor of anchors) {
            const nodeAssociation = await this.repository.association.findByObjUid(anchor.uid);
            if (nodeAssociation) nodeAssociations.push(nodeAssociation);
        }

        return { sublocation, anchors, nodeAssociations };
    }

    /**
     * @method
     * @param {Object} params - options for read sublocation in repository
     * @param {String} params.uid - sublocation unique identifier
     * @return {Promise<Object>}
     */
    async remove(params) {
        const { uid } = params;
        const sublocation = await this.repository.sublocation.findByUid(uid);
        if (!sublocation) {
            throw new NotFoundError(`Sublocation with uid ${uid} not found`);
        }
        await this.repository.sublocation.deleteByUid(uid);
        await new SublocationId().remove(sublocation.id);

        return sublocation;
    }

    /**
     * @method
     * @param {Object} params - sublocation properties to update
     * @param {String} params.uid - sublocation unique identifier
     * @param {String|undefined} params.name - sublocation name
     * @return {Promise<Object>}
     */
    async update(params) {
        const { uid } = params;
        const sublocation = await this.repository.sublocation.findByUid(uid);
        if (!sublocation) {
            throw new NotFoundError(`Sublocation with uid ${uid} not found`);
        }

        await this.repository.sublocation.save({ ...params, id: sublocation.id, updated_at: new Date().toISOString() });

        return formatSublocationResponse(await this.repository.sublocation.findByUid(uid));
    }
}
