import ModelBase from '../ModelBase.mjs';
import { NODE_TYPE } from '../../system/constants.mjs';
import { NotFoundError, ValidationError } from '../errors/index.mjs';

/**
 * @swagger
 * components:
 *  schemas:
 *      TagAssignmentCreateBody:
 *          type: object
 *          properties:
 *              mac:
 *                  type: string
 *                  required: true
 *              trackable_uid:
 *                  type: string
 *                  required: true
 *      TagAssignmentCloseBody:
 *          type: object
 *          properties:
 *              mac:
 *                  type: string
 *                  required: false
 *              trackable_uid:
 *                  type: string
 *                  required: false
 *      TagAssociation:
 *          type: object
 *          properties:
 *              trackable_uid:
 *                  type: string
 *                  description: The unique identifier of the assigned trackable.
 *              mac:
 *                  type: string
 *                  description: The MAC address of the assigned node. In upper case.
 *              created_at:
 *                  type: string (ISO 8601)
 *                  description: The time stamp when the node association was created, in ISO 8601 format.
 *      TagAssociations:
 *          type: object
 *          properties:
 *              tag_assignments:
 *                  type: array
 *                  items:
 *                      $ref: '#/components/schemas/TagAssociation'
 */

export class TagAssociation extends ModelBase {
    #entityType = NODE_TYPE.TO;

    /**
     * @property {String|null}
     */
    static object = 'tag_assignment';

    /**
     * @method
     * @param {Object} params - association properties to save
     * @param {String} params.mac - tag mac address
     * @param {String} params.trackable_uid - trackable unique identifier
     * @return {Promise<Object>}
     */
    async save(params) {
        const now = new Date().toISOString();
        const trackable = await this.repository.trackable.findByUid(params.trackable_uid);

        if (!trackable) {
            throw new NotFoundError(`Trackable with uid ${params.trackable_uid} not found`);
        }

        const node = await this.repository.node.findByMac(params.mac);

        if (!node || node.type !== this.#entityType) {
            throw new NotFoundError(`Tag with mac ${params.mac} not found`);
        }

        const associationWithTag = await this.repository.association.findByMac(params.mac);

        if (associationWithTag) {
            throw new ValidationError(`Assignment with tag ${params.mac} already exist`);
        }

        const associationWithTrackable = await this.repository.association.findByObjUid(params.trackable_uid);

        if (associationWithTrackable) {
            throw new ValidationError(`Assignment with trackable ${params.trackable_uid} already exist`);
        }

        const tagAssociation = {
            created_at: now,
            ...params,
        };

        return this.repository.association.save(tagAssociation);
    }

    /**
     * @method
     * @param {Object} params - association properties
     * @param {String|undefined} params.uid - trackable unique identifier
     * @return {Promise<Object>}
     */
    async load(params = {}) {
        if (params.uid) {
            return this.repository.association.findByObjUid(params.uid);
        }
        return ((await this.repository.association.findAll()) || []).filter((association) => association.trackable_uid);
    }

    /**
     * @method
     * @param {Object} params - association properties
     * @param {String} params.mac - tag mac address
     * @param {String} params.trackable_uid - unique id of trackable
     * @return {Promise<Object>}
     */
    async close(params) {
        const { mac, trackable_uid } = params;
        if (!mac && !trackable_uid) {
            throw new ValidationError(`"mac" or "trackable_uid" require`);
        }
        let association;
        if (mac && trackable_uid) {
            association = await this.repository.association.findByMac(mac);
            if (!association || (association && association.trackable_uid !== trackable_uid)) {
                throw new NotFoundError(`Assignment with tag ${mac} and trackable ${trackable_uid} not found`);
            }
        } else if (mac) {
            association = await this.repository.association.findByMac(mac);
            if (!association) {
                throw new NotFoundError(`Assignment with tag ${mac} not found`);
            }
        } else if (trackable_uid) {
            association = await this.repository.association.findByObjUid(trackable_uid);
            if (!association) {
                throw new NotFoundError(`Assignment with trackable ${trackable_uid} not found`);
            }
        }

        await this.repository.association.del(association);

        return association;
    }
}
