import flat from 'flat';
import RepoBase from './RepoBase.mjs';
import { RepoError } from './models/index.mjs';

const flatten = flat.flatten;
const unflatten = flat.unflatten;

export default class RepoRedisBase extends RepoBase {
    /**
     * @typedef {Class} RepoRedisBase
     */

    /**
     * @param {Object} redisClient - DB client
     * @param {String} document - document name
     * @param {Number} database - database number, by default equal to 0
     */
    constructor(redisClient, document, database = 0) {
        super(redisClient, document);
        console.log('Repository document: ', document);

        this.database = database;
    }

    /**
     * @method
     * @param {String} key - DB key for search in DB
     * @param {Boolean} unflat - flag that determine necessity of use flat operation for returned result
     * @return {Promise<Object>} returns DB data
     */
    async getByKey(key, unflat = true) {
        const flatten_obj = await this.repoInstance.hgetall(key);
        if (!flatten_obj) {
            return null;
        }
        if (unflat) {
            return unflatten(flatten_obj);
        }
        return flatten_obj;
    }

    /**
     * @method
     * @param {String<22>} uid - entity unique identifier
     * @return {Promise<Object>} returns result of key removing
     */
    async deleteByUid(uid) {
        return this.repoInstance.del(`${this.document}:${uid}`);
    }

    /**
     * @method
     * @param {String<22>} uid - entity unique identifier
     * @return {Promise<Object>} returns result of search by key
     */
    async findByUid(uid) {
        const flatten_obj = await this.repoInstance.hgetall(`${this.document}:${uid}`);
        if (!flatten_obj) {
            return null;
        }
        return unflatten(flatten_obj);
    }

    /**
     * @method
     * @return {Promise<Array<Object>>} returns entities list
     */
    async findAll() {
        const result = [];
        try {
            const keys = await this.scanAll('0', `${this.document}:*`, new Set());
            // console.log('keys: ', keys);
            for (const key of keys) {
                const obj = await this.getByKey(key);
                // console.log('obj: ', obj);
                if (obj) {
                    result.push(obj);
                }
            }
            return result;
        } catch (e) {
            throw new RepoError(e);
        }
    }

    /**
     * @method
     * @param {Object} cursor - ???
     * @param {String} pattern - ???
     * @param {Object} returnSet - ???
     * @return {Promise<Array<Object>>} returns result of search by key
     */
    async scanAll(cursor, pattern, returnSet) {
        const reply = await this.repoInstance.scan(cursor, 'MATCH', `${pattern}`, 'COUNT', 10000);
        cursor = reply[0];
        const keys = reply[1];
        if (!keys) {
            return returnSet;
        }

        // eslint-disable-next-line no-unused-vars
        keys.forEach(function (key, i) {
            returnSet.add(key);
        });

        if (cursor === '0') {
            return Array.from(returnSet);
        } else {
            return await this.scanAll(cursor, pattern, returnSet);
        }
    }

    /**
     * @method
     * @param {Object} obj - data to save in DB
     * @return {Promise<Object>} returns data saved in DB
     */
    async save(obj) {
        if (!this.document) {
            throw new RepoError(`save: no document`);
        }
        if (!obj || !obj.uid || typeof obj !== 'object') {
            throw new RepoError(`save: object saving is aborted due to object format: ${JSON.stringify(obj)}`);
        }
        const fields = [];
        for (const [key, value] of Object.entries(flatten(obj))) {
            if (value || typeof value === 'number') {
                // do not save empty values
                fields.push(key, value);
            }
        }

        await this.repoInstance.hmset(`${this.document}:${obj.uid}`, ...fields);
        return obj;
    }

    /**
     *
     * @param {Object} obj
     * @property {String<String>} obj.uid
     * @returns {Promise<void>}
     */
    async del(obj) {
        if (!this.document) {
            throw new RepoError(`save: no document`);
        }
        if (!obj || !obj.uid || typeof obj !== 'object') {
            throw new RepoError(`del: object removal is aborted due to object format: ${obj}`);
        }

        await this.repoInstance.del(`${this.document}:${obj.uid}`);
    }

    transaction() {
        return this.repoInstance.client.multi();
    }
    exec(transaction) {
        return transaction.exec();
    }
}
