import flat from 'flat';
import RepoRedisBase from '../RepoRedisBase.mjs';
import { RepoError } from '../models/index.mjs';

const flatten = flat.flatten;
const unflatten = flat.unflatten;

/**
 * @typedef PlatformSublocation {Object}
 * @property uid {String}
 * @property name {String}
 * @property created_at {String} ISO date
 * @property updated_at {String} ISO date
 * @property id {Number}
 * @property location_uid {String}
 * @property elevation {Number}
 */

export class SublocationRepository extends RepoRedisBase {
    /**
     * @typedef {Class} SublocationRepository
     */

    /**
     * @param {Object} options - repository options
     * @param {Object} options.repositoryInstance - database connection
     * @param {String} options.prefix - database prefix
     */
    constructor(options) {
        const { prefix, repositoryInstance } = options;

        super(repositoryInstance, `${prefix}:sublocation`);
    }

    /**
     * @method
     * @param {String} uid - entity unique identifier
     * @return {Promise<Object>} returns result of search by key
     */
    async findByUid(uid) {
        const flatten_obj = await this.repoInstance.hgetall(`${this.document}:uid:${uid}`);
        if (!flatten_obj) {
            return null;
        }
        return unflatten(flatten_obj);
    }

    /**
     * @method
     * @return {Promise<Array<Object>>} returns entities list
     */
    async findAll() {
        const result = [];
        try {
            const keys = await this.scanAll('0', `${this.document}:uid:*`, new Set());
            // logger.info('keys: ', keys);
            for (const key of keys) {
                const obj = await this.getByKey(key);
                // logger.info('obj: ', obj);
                if (obj) {
                    result.push(obj);
                }
            }
            return result;
        } catch (e) {
            throw new RepoError(e);
        }
    }

    /**
     * @method
     * @param {Object} obj - data to save in DB
     * @return {Promise<Object>} returns data saved in DB
     */
    async save(obj) {
        if (!this.document) {
            throw new RepoError(`save: no document`);
        }
        if (!obj || !obj.uid || !obj.id || typeof obj !== 'object') {
            throw new RepoError(`save: object saving is aborted due to object format: ${JSON.stringify(obj)}`);
        }
        const fields = [];
        for (const [key, value] of Object.entries(flatten(obj))) {
            if (value || typeof value === 'number') {
                // do not save empty values
                fields.push(key, value);
            }
        }

        await this.repoInstance.hmset(`${this.document}:uid:${obj.uid}`, ...fields);
        await this.repoInstance.set(`${this.document}:id:${obj.id}`, obj.uid);
        return obj;
    }

    /**
     * @method
     * @param {String} uid - entity unique identifier
     * @return {Promise<Object>} returns result of key removing
     */
    async deleteByUid(uid) {
        const sublocation = await this.repoInstance.hgetall(`${this.document}:uid:${uid}`);
        if (!sublocation) {
            throw new RepoError(`delete: no document`);
        }
        this.repoInstance.client
            .multi()
            .del(`${this.document}:uid:${sublocation.uid}`)
            .del(`${this.document}:id:${sublocation.id}`)
            .exec();
        return uid;
    }

    /**
     * @method
     * @param {Number} id - sublocation numeric identifier
     * @return {Promise<String>} returns sublocation uid
     */
    async findById(id) {
        const uid = await this.repoInstance.get(`${this.document}:id:${id}`);
        const flatten_obj = await this.repoInstance.hgetall(`${this.document}:uid:${uid}`);
        if (!flatten_obj) {
            return null;
        }
        return unflatten(flatten_obj);
    }

    /**
     * @method
     * @param {Number} id - sublocation numeric identifier
     * @return {Promise<String>} returns sublocation uid
     */
    async findUidById(id) {
        const uid = await this.repoInstance.get(`${this.document}:id:${id}`);
        if (!uid) {
            return null;
        }
        return uid;
    }
}
