import flat from 'flat';
import RepoRedisBase from '../RepoRedisBase.mjs';
import { RepoError } from '../models/index.mjs';

const unflatten = flat.unflatten;

/**
 * @typedef PlatformTrackable {Object}
 * @property uid {String}
 * @property name {String}
 * @property tag_mount_height {Number}
 * @property heartbeat {String}
 * @property position_reporting.auto {Number}
 * @property position_reporting.interval {Number}
 * @property created_at {String}
 * @property updated_at {String}
 * @property position {Object}
 * @property position.x {Number}
 * @property position.y {Number}
 * @property position.z {Number}
 */

export class TrackableRepository extends RepoRedisBase {
    assocResource;
    /**
     * @typedef {Class} TrackableRepository
     */

    /**
     * @param {Object} options - repository options
     * @param {Object} options.repositoryInstance - database connection
     * @param {String} options.prefix - database prefix
     */
    constructor(options) {
        const { prefix, repositoryInstance } = options;
        super(repositoryInstance, `${prefix}:tko`);
        this.assocResource = `${prefix}:locks:assoc`;
    }

    /**
     * @method
     * @param {String} uid - trackable uid
     * @param {Object} position - JSON object of the trackable position
     * @param {Number} position.x - X coordinate of trackable position
     * @param {Number} position.y - Y coordinate of trackable position
     * @param {Number} position.z - Z coordinate of trackable position
     * @param {Number} position.sublocation_uid - sublocation unique identifier
     * @return {Promise<Object>} returns result of position saving
     */
    async savePosition(uid, position) {
        if (!uid) {
            throw new RepoError('specify uid');
        }
        if (!position || typeof position !== 'object') {
            throw new RepoError('specify position as object');
        }
        return await this.repoInstance.hmset(`${this.document}:${uid}`, 'position', JSON.stringify(position));
    }

    /**
     *
     * @param obj {PlatformTrackable}
     * @returns {Promise<{position}|Object|Object>}
     */
    async save(obj) {
        if (!obj.position) {
            return super.save(obj);
        } else if (typeof obj.position === 'object') {
            const tko = { ...obj,
                position: JSON.stringify(obj.position)
            }
            await super.save(tko);
            return obj;

        } else if (typeof obj.position === 'string') {
            return super.save(obj);
        }


    }
    deflate(flatObject) {
        if (flatObject.position) {
            const position = JSON.parse(flatObject.position);
            delete flatObject.position;
            return {
                position,
                ...unflatten(flatObject),
            };
        } else {
            return unflatten(flatObject);
        }
    }
    /**
     * @method
     * @return {Promise<Array<Object>>} returns entities list
     */
    async findAll() {
        return (await super.findAll() || []).map(flatObject => this.deflate(flatObject));
    }
    /**
     * @method
     * @param {String} uid - trackable uid
     * @return {Promise<Object>} returns trackable
     */
    async findByUid(uid) {
        if (!uid) {
            throw new RepoError('specify uid in findByUid method');
        }
        const flatObject = await this.repoInstance.hgetall(`${this.document}:${uid}`);
        if (!flatObject) {
            return null;
        }
        return this.deflate(flatObject)
    }
    /**
     * @method
     * @param {String<22>} uid - entity unique identifier
     * @return {Promise<Object>} returns result of key removing
     */
    async deleteByUid(uid) {
        const prefix = this.document.split(':')[0];
        const objAssocKey = `${prefix}:assoc:obj_uid:${uid}`;
        const tkoKey = `${this.document}:${uid}`;
        const client = this.repoInstance.client;
        return this.repoInstance.executeWithLock(this.assocResource,async () => {
            const assoc = await this.repoInstance.hgetall(objAssocKey);
            if (assoc) {
                const macAssocKey = `${prefix}:assoc:mac:${assoc.mac}`;
                client.multi().del(macAssocKey).del(objAssocKey).del(tkoKey).exec();
                return 1;
            } else {
                return await this.repoInstance.del(tkoKey);
            }
        });
    }
}
